import obspy
from obspy import UTCDateTime
from scipy import interpolate
import numpy as np
import os
from seisloc.utils import month_day

def calculate_snr(signal, noise):
    """
    Calculate the signal-to-noise ratio (SNR) of a given signal and noise.
    
    Parameters:
        signal (numpy.ndarray): Array representing the signal.
        noise (numpy.ndarray): Array representing the noise.
    
    Returns:
        float: The signal-to-noise ratio (SNR) in decibels (dB).
    """
    signal_power = np.mean(np.abs(signal) ** 2)
    noise_power = np.mean(np.abs(noise) ** 2)
    snr = 10 * np.log10(signal_power / noise_power)
    return snr

def tr_snr(tr,noiseLen=0.5,sigLen=0.5,shift=0,mkr='a'):
    """
    Calculate the signal-to-noise ratio (SNR) of a given trace.
    
    Parameters:
    |    tr (obspy.Trace): The trace to calculate the SNR for.
    |    noiseLen, sigLen: noise window length and signal window length wrt. mkr
    |    shift: separation between signal and noise will be t(mkr)+shift
    |    mkr: the marker in the sac file
    
    Returns:
    |    float: The signal-to-noise ratio (SNR) in decibels (dB).
    """
    starttime = tr.stats.starttime

    delta = tr.stats.delta
    b = tr.stats.sac.b
    mkrVal = getattr(tr.stats.sac,mkr)
    signalBeginTob = mkrVal-b+shift # separation points w.r.t the b value
    idxSignal = int(signalBeginTob/delta)
    idxNoise = int((mkrVal-b-noiseLen+shift)/delta)
    signal = tr.data[idxSignal:idxSignal+int(sigLen/delta)]
    noise = tr.data[idxNoise:idxSignal]
    
    # Calculate the SNR
    snr = calculate_snr(signal, noise)
    return snr

def get_trav(st,sta,marker="a"):
    """
    Get the marker time from SAC header
    if marker is a string, then direct get the marker time. The absence 
    of the corresponding marker time will lead to the return of "0, 'none'".

    If marker is a list, then will sequential process marker one by one and 
    will return upon the successful load of the travel time. 

    For example, for a dataset of events, the P arrivals are first picked with 
    marker 'a'. After that, some P arrivals are refined with marker "t3" and 
    further refined with marker 't5'. Then to get the best travel time, one can
    use:
    >>> get_trav(st,sta,marker=['t5','t3','a'])

    Return:
    marker_time, marker
    """
    status = False   # whether load travel time successfully
    st=st.select(station=sta)
    if len(st)==0:
        raise Exception("No data for station "+sta)
    if isinstance(marker,str):
        if hasattr(st[0].stats.sac,marker):
            trav=getattr(st[0].stats.sac,marker)
            status = True

    elif isinstance(marker,list):
        markers = marker
        for marker in markers:
            if hasattr(st[0].stats.sac,marker):
                trav=getattr(st[0].stats.sac,marker)
                status = True
                break
    else:
        raise Exception("marker should be either a 'str' or a 'list'")

    if status == False:
        trav=0
        marker="none"

    return trav,marker

def aligned_sac_datas(tr1,tr2,scc_dt,tb,te,marker='t0'):
    """
    Parameters:
       tr1,tr2: obspy trace
        scc_dt: time difference measured by cross-correlation, t1-t2
            tb: Waveform cut time before the S arrival
            te: Waveform cut time after the S arrival
        marker: 'a' for P arrival time, 't0' for S arrival time
    """
    tr1.detrend("linear")
    tr1.detrend("constant")
    tr2.detrend("linear")
    tr2.detrend("constant")
    
    t01 = getattr(tr1.stats.sac,marker)
    t02 = getattr(tr2.stats.sac,marker)
    b1 = tr1.stats.sac.b
    b2 = tr2.stats.sac.b
    trim_b1 = tr1.stats.starttime+(t01-b1)+scc_dt+tb
    trim_e1 = trim_b1+(te-tb)

    trim_b2 = tr2.stats.starttime+(t01-b2)+tb
    trim_e2 = trim_b2+(te-tb)

    tr1.trim(starttime = trim_b1, endtime = trim_e1)
    tr2.trim(starttime = trim_b2, endtime = trim_e2)
    
    tr1.detrend("constant")
    tr1.detrend("linear")
    tr2.detrend("constant")
    tr2.detrend("linear")

    return tr1.data,tr2.data

def sac_interpolation(insacPth,outsacPth,factor=10):
    """
    Interpolate waveform data using the quadratic fitting method provided by scipy.interpolate.interp1d module
    """
    insacPth = os.path.abspath(insacPth)
    st = obspy.read(insacPth)
    data = st[0].data
    delta = st[0].stats.delta
    xs = np.arange(len(data))
    f = interpolate.interp1d(xs,data,kind='quadratic')
    
    deltaNew = delta/factor
    ixs = np.arange((len(data)-1)*factor+1)*1/factor
    dataInterp = f(ixs)
    stInterp = st.copy()
    stInterp[0].data = dataInterp
    stInterp[0].stats.delta = deltaNew
    stInterp[0].write(outsacPth,format="SAC")
    
def read_sac_ref_time(tr):
    """
    Read and return reference time of a sac file in obspy.UTCDateTime format.

    Parameter
    --------
    tr: Trace object of obspy
    """

    nzyear = tr.stats.sac.nzyear
    nzjday = tr.stats.sac.nzjday
    nzhour = tr.stats.sac.nzhour
    nzmin = tr.stats.sac.nzmin
    nzsec = tr.stats.sac.nzsec
    nzmsec = tr.stats.sac.nzmsec*0.001
    year,month,day = month_day(nzyear,nzjday)
    sac_ref_time = UTCDateTime(year,month,day,nzhour,nzmin,nzsec)+nzmsec
    return sac_ref_time


def get_tr_marker_idx(tr,marker='a'):
    refTime = read_sac_ref_time(tr)
    sp = tr.stats.sampling_rate
    starttime = tr.stats.starttime
    markerTime = refTime+tr.stats.sac[marker]
    idx = np.round(int((markerTime - starttime)*sp),0)
    return idx

def get_tr_data(tr,idx,tb,te):
    """
    tb,te: negative/positive for waveform before/after the marker
    
    """
    delta = tr.stats.delta
    shift1 = int(tb/delta)
    shift2 = int(te/delta)
    data = tr.data[idx+shift1:idx+shift2]
    return data



from seisloc.math import lsfit
import numpy as np
from numba import jit,prange
from numpy.linalg import det
from obspy.geodetics import gps2dist_azimuth

from numba import jit, prange
import numpy as np

@jit(nopython=True,fastmath=True)
def grid_search(G:np.ndarray,dts:np.ndarray,n2x:int,dx:float,n2y:int,dy:float,n2z:int,dz:float,W=1,secondRefine:bool=False)->tuple:
    """
    n2x: number of grid points in the positive x direction
    """
    nx = 2*n2x+1
    ny = 2*n2y+1
    nz = 2*n2z+1
    xs = np.arange(-n2x,n2x+1,1)*dx
    ys = np.arange(-n2y,n2y+1,1)*dy
    zs = np.arange(-n2z,n2z+1,1)*dz
    resMatrix = np.zeros((nx,ny,nz))
    for i in prange(nx):
        for j in prange(ny):
            for z in prange(nz):
                m = np.array([[xs[i]],[ys[j]],[zs[z]]])
                ets = G@m
                resMatrix[i,j,z] = np.std((ets-dts)*W)
    minRes = np.min(np.abs(resMatrix))
    k = np.where(resMatrix==minRes)

    # Initial best fit
    x0 = xs[k[0][0]]
    y0 = ys[k[1][0]]
    z0 = zs[k[2][0]]
    # check whether local minium
    if x0==xs[0] or x0==xs[-1] or\
       y0==ys[0] or y0==ys[-1] or\
       z0==zs[0] or z0==zs[-1]:
       locMin = 0
    else:
       locMin = 1

    #-------- further refine -------------------
    if secondRefine:
        xs1 = xs/10
        ys1 = ys/10
        zs1 = zs/10
        
        resMatrix = np.zeros((nx,ny,nz))
        for i in prange(nx):
            for j in prange(ny):
                for z in prange(nz):
                    m = np.array([[x0+xs1[i]],[y0+ys1[j]],[z0+zs1[z]]])
                    ets = G@m
                    resMatrix[i,j,z] = np.std((ets-dts)*W)
        
        minRes = np.min(np.abs(resMatrix))
        k = np.where(resMatrix == minRes)

        x0 = x0+xs1[k[0][0]]
        y0 = y0+ys1[k[1][0]]
        z0 = z0+zs1[k[2][0]]

    return x0,y0,z0,locMin

def locate_slave2(masLon,masLat,masDep,masVels,
                 phaseRecs,model,stas,stepDict,slaveId=-1,secondRefine=False):
    """
    Relative location of the template matching detected events with respect to 
    the template event (master event) using the grid-search method.

    Parameters
    | masLon,masLat,masDep: master event parameters
    |   masVels: P ans S velocties of the master event position
    | phaseRecs: Records (sta,dt,weight,phsType) of phase relationships between 
                 the slave event and the master event. dt = t(sla.) - t(mas.)
    |     model: Obspy taup model for the ray-tracing
    |  stepDict: Information regarding steps. E.g., {"x":[5,0.1];"y":[5,0.1],"z":{5,0.1}}
                 stands for search along x,y,z for -0.5 to 0.5 km with step length of 0.1 km.
    |      stas: seisloc.sta Sta() class that containing stations information
    |   slaveId: This parameter is designed for multiprocessing and will be 
                 returned as a marker of the multiprocessing result

    [[dtdx1,dtdy1,dtdz1],
     [dtdx2,dtdy2,dtdz2],
     [dtdx3,dtdy3,dtdz3],
     ...
     dtdxn,dtdyn,dtdzn]] @ [[ix],[iy],[iz]] = [[et1],[et2],[et3],...,[etn]] (et, estimated time)

    best location is the np.min(np.std([[et1],[et2],[et3],...,[etn]]-[[dt1],[dt2],[dt3],...,[dtn]]))

    """
    G = [] # data kernel 
    d = [] # observation 
    W = [] # weights
    # ----- set up kernel -----------
    for i,phaseRec in enumerate(phaseRecs):
        sta,dt,weight,phsType = phaseRec
        staLon = stas.dict["SC"][sta][0]
        staLat = stas.dict["SC"][sta][1]
        dist,az,baz = gps2dist_azimuth(masLat,masLon,staLat,staLon)
        distKm = dist/1000
        distDeg = distKm/111.19
        if phsType.lower() == "p":
            masVel = masVels[0]
            weight = 0.5
        if phsType.lower() == "s":
            masVel = masVels[1]
            weight = 1
        phase_list = [phsType.lower(),phsType.upper()]
        arrivals = model.get_ray_paths(source_depth_in_km=masDep,
                                       distance_in_degree=distDeg,
                                       phase_list = phase_list)
        

        if len(arrivals)>0:

            takeAngle = arrivals[0].takeoff_angle
            phkm = np.sin(np.deg2rad(180-takeAngle))/masVel
            dtdz = -phkm / np.tan(np.deg2rad(takeAngle))
            dtdx = -phkm * np.sin(np.deg2rad(az))
            dtdy = -phkm * np.cos(np.deg2rad(az))

            G.append([dtdx,dtdy,dtdz])
            d.append([dt])
            W.append(weight)
    G = np.array(G)
    d = np.array(d).reshape((len(d),1))
    W = np.array(W).reshape((len(W),1))
    
    xnode=stepDict["x"][0]
    xstep=stepDict["x"][1] # -xnode*xstep, ...,0,...,xnode*step
    ynode=stepDict["y"][0]
    ystep=stepDict["y"][1] # -xnode*xstep, ...,0,...,xnode*step
    znode=stepDict["z"][0]
    zstep=stepDict["z"][1] # -xnode*xstep, ...,0,...,xnode*step
    dxkm,dykm,dzkm,bdyStats = grid_search(G,d,xnode,xstep,ynode,ystep,znode,zstep,W=W,secondRefine=secondRefine)

    dlon = dxkm/(111.19*np.cos(np.deg2rad(masLat)))
    dlat = dykm/111.19

    return masLon+dlon,masLat+dlat,masDep+dzkm,bdyStats,slaveId

def locate_slave(masLon,masLat,masDep,masVels,
                 phaseRecs,model,stas,slaveId=-1):
    """
    Relative location of the template matching detected events with respect to 
    the template event (master event)

    Parameters
    | masLon,masLat,masDep: master event parameters
    |   masVels: P ans S velocties of the master event position
    | phaseRecs: Records (sta,dt,weight,phsType) of phase relationships between 
                 the slave event and the master event. dt = t(sla.) - t(mas.)
    |     model: Obspy taup model for the ray-tracing
    |      stas: seisloc.sta Sta() class that containing stations information
    |   slaveId: This parameter is designed for multiprocessing and will be 
                 returned as a marker of the multiprocessing result

    """
    G = [] # data kernel 
    d = [] # observation 
    W = [] # weights
    # ----- set up kernel -----------
    for i,phaseRec in enumerate(phaseRecs):
        sta,dt,weight,phsType = phaseRec
        staLon = stas.dict["SC"][sta][0]
        staLat = stas.dict["SC"][sta][1]
        dist,az,baz = gps2dist_azimuth(masLat,masLon,staLat,staLon)
        distKm = dist/1000
        distDeg = distKm/111.19
        if phsType.lower() == "p":
            masVel = masVels[0]
        if phsType.lower() == "s":
            masVel = masVels[1]
        phase_list = [phsType.lower(),phsType.upper()]
        arrivals = model.get_ray_paths(source_depth_in_km=masDep,
                                       distance_in_degree=distDeg,
                                       phase_list = phase_list)
        if len(arrivals)>0:
            takeAngle = arrivals[0].takeoff_angle
            phkm = np.sin(np.deg2rad(180-takeAngle))/masVel
            dtdz = -phkm / np.tan(np.deg2rad(takeAngle))
            dtdx = -phkm * np.sin(np.deg2rad(az))
            dtdy = -phkm * np.cos(np.deg2rad(az))
            G.append([dtdx,dtdy,dtdz])
            d.append([dt])
            W.append(weight)
    G = np.array(G)
    # ----- Normalize G columns to enhance the stability of inversion -----
    norm0 = np.linalg.norm(G[:,0])
    norm1 = np.linalg.norm(G[:,1])
    norm2 = np.linalg.norm(G[:,2])
    G[:,0] = G[:,0]/norm0
    G[:,1] = G[:,1]/norm1
    G[:,2] = G[:,2]/norm2
    d = np.array(d)
    W = np.diag(W)
    WG = np.matmul(W,G)
    GTWG = np.matmul(G.T,WG)
    if len(G)<3 or det(GTWG) == 0: # No enough observation or singular matrix
        m = np.zeros((3,1))
    else:
        m = lsfit(G,d,W)
    # ----- G@m=d --> G/norm @ norm*m = d -----
    dxkm = m.ravel()[0]/norm0
    dykm = m.ravel()[1]/norm1
    dzkm = m.ravel()[2]/norm2
    dlon = dxkm/(111.19*np.cos(np.deg2rad(masLat)))
    dlat = dykm/111.19
                            
    return masLon+dlon,masLat+dlat,masDep+dzkm,slaveId



import os
from obspy.geodetics import gps2dist_azimuth

def geo_spatial_dist(lo1,la1,dp1,lo2,la2,dp2):
    """
    Input depth in kilometers
    Return distance in kilometers
    """
    distHm,_,_ = gps2dist_azimuth(la1,lo1,la2,lo2)
    distHkm = distHm/1000
    ddep = dp1 - dp2

    return (distHkm**2+ddep**2)**0.5



import os
import copy
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from obspy import UTCDateTime
from obspy.geodetics import gps2dist_azimuth
from seisloc.dd import load_DD
from seisloc.hypoinv import Hypoinv
from seisloc.geometry import in_rectangle,loc_by_width
from math import floor,ceil
from seisloc.statistics import sum_count_Mo,write_sum_count_Mo
from seisloc.phase_convert import cata2fdsn
from tqdm import tqdm

def _load_cata(locFile,format="hypoDD"):
    if format == "hypoDD":
        tmpDict, _ = load_DD(locFile)
        return tmpDict
    if format == "sum": # Hypoinverse summary file
        from seisloc.hypoinv import load_sum_evid
        sumDict = load_sum_evid(locFile)
        tmpDict = {}
        for evid in sumDict.keys():
            estr,evlo,evla,evdp,emag,eres = sumDict[evid]
            etime = UTCDateTime.strptime(estr,"%Y%m%d%H%M%S%f")
            tmpDict[evid] = [evlo,evla,evdp,emag,etime]
        return tmpDict
    if format == "cata":
        with open(locFile,'r') as f:
            tmpDict={}
            for line in f:
                line = line.strip()
                _evid,_evlo,_evla,_evdp,_emag,_eday,_etime=line.split()
                tmpDict[int(_evid)] = [float(_evlo),float(_evla),float(_evdp),float(_emag),UTCDateTime(_etime)]
        return tmpDict



class Catalog():
    def __init__(self,locFile="hypoDD.reloc",format="hypoDD",verbose=1):
        """
        The programme will read in catalog in format ("hypoDD", "sum", and "cata"). If no hypoDD
        file provided (locFile=None), it will generate an empty catalog. 
        A user can set up a new catalog by providing a dict in the form:
            dict[evid] = [lon,lat,dep,mag,UTCDateTime]
        example:
        >>> cata = Catalog(locFile=None)
        >>> cata.dict = cataDict  # cata_dict is a dictionary follows above format
        >>> cata.init()            # initiation of the class
        >>> print(cata)            # basic information will be printed
        """
        self.verbose = verbose
        if locFile != None:
            if not os.path.exists(locFile):
                raise Exception(f"{locFile} not existed!")
            self.dict = _load_cata(locFile,format=format)
            if self.verbose==True or self.verbose==1:
                print("[Class Catalog] successfully loaded the catalog file: "+locFile)
            self.init()
        else:
            if self.verbose == True or self.verbose==1:
                print("[Class Catalog] ~~~~~~~~~~~~~~~ Warning ~~~~~~~~~~~~~~~~~")
                print("[Class Catalog] No hypoDD .reloc file provided, an empty Catalog created.")
                print("[Class Catalog] You can define self.dict[evid] = [lon,lat,dep,mag,UTCDateTime]}")
                print("[Class Catalog] Then run: .init() to initiate the catalog.")
            self.dict = {}

    def init(self):
        """
        Initiate the catalog
        """
        if self.verbose == True or self.verbose==1:
            print("[Class Catalog] Initiate the catalog ... ")
        self.baseTime = UTCDateTime(2020,1,1)
        self.init_keys()
        self.init_locs()
        self.init_relative_seconds()
        self.sort()
        self.evids = self.keys
        if self.verbose == True or self.verbose==1:
            print("[Class Catalog] Initiation completed! ")

    def init_keys(self):
        """
        Build up event ids array
        """
        self.keys = list(self.dict.keys())
        self.keys = np.array(self.keys)
        self.keys = self.keys.astype(int)

    def init_locs(self):
        """
        Generate numpy array in format lon, lat, dep, mag
        """
        self.locs = []
        for key in self.keys:
            lon = self.dict[key][0]
            lat = self.dict[key][1]
            dep = self.dict[key][2]
            mag = self.dict[key][3]
            self.locs.append([lon,lat,dep,mag])
        self.locs = np.array(self.locs)

    def init_relative_seconds(self):
        """
        Numpy array to save relative seconds to the first event
        """
        self.relative_seconds = []
        for key in self.keys:
            etime = self.dict[key][4]
            self.relative_seconds.append(etime-self.baseTime)
        self.relative_seconds = np.array(self.relative_seconds)

    def update_keys(self,idxs):
        """
        Update keys array with indexs
        """
        self.keys = self.keys[idxs]

    def update_dict(self):
        """
        Update dictionary with new set of keys
        """
        old_keys = list(self.dict.keys())
        for key in old_keys:
            if key not in self.keys:
                self.dict.pop(key)

    def update_locs(self,idxs):
        """
        Update location array with indexs
        """
        self.locs = self.locs[idxs]

    def update_relative_seconds(self,idxs):
        """
        Update relative times with indexs
        """
        self.relative_seconds = self.relative_seconds[idxs]
        
    def crop(self,lonmin,lonmax,latmin,latmax):
        """
        Trim the dataset with the lon-lat boundary conditions
        """
        idxs = np.where((self.locs[:,0]>=lonmin)&(self.locs[:,0]<=lonmax)&\
                        (self.locs[:,1]>=latmin)&(self.locs[:,1]<=latmax))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def magsel(self,mag_low,mag_top=10):
        """
        Select the dataset with the magnitude
        """
        idxs = np.where((self.locs[:,3]>=mag_low)&(self.locs[:,3]<=mag_top))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def trim(self,starttime,endtime):
        """
        Trim the dataset with time conditions
        """
        min_reftime = starttime - self.baseTime
        max_reftime = endtime - self.baseTime
        
        idxs = np.where((self.relative_seconds>=min_reftime)&\
                        (self.relative_seconds<=max_reftime))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def sort(self,method="time"):
        idxs = self.relative_seconds.argsort()
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def hplot(self,
              xlim=[],
              ylim=[],
              figSize=None,
              edgeColor='grey',
              edgeWidth=0.5,
              markerSize=6,
              sizeRatio=1,
              impMag=None,
              refTime = None,
              cmap = None,
              vmin=0,
              vmax=1,
              unit="day",
              addSection=False,
              alonlat=[104,29],
              blonlat=[105,30],
              secWidth=0.05,
              crop=False,
              ax = None):
        """
        Map view plot of earthquakes,earthquake denoted default by black circle
        Parameters:
        |         xlim: longitude limit, e.g. [104,105]
        |         ylim: latitude limit, e.g. [29,30]
        |      figSize: e.g. (5,5)
        |    edgeColor: earthquake marker(circle) edgeColor
        |      impMag: important magnitude. Magnitude larger than this level will be 
        |               highlighted
        |     refTime: reference time in UTCDateTime used to constrain colormap, if
        |               no colormap provided, seismicity will plotted by default
        |         cmap: colormap, check 'matplotlib' for more detail.
        |    vmin,vmax: the minimum and maximum value for colormap
        |         unit: "day","hour", or "second" for vmin and vmax
        |  addSection: if want to add one cross-section, set True
        |      alonlat: the [lon,lat] of the section start point 'a'
        |      blonlat: the [lon,lat] of the section end point 'b'
        |secWidth: width of section in degree
        |         crop: if True, the dataset will cut dataset to leave only events
                        inside the cross-section
        """
        #---------------- initiation ----------------------------------
        if ax != None:     # get current axis
            plt.sca(ax)
        else:
            if figSize != None:
                plt.figure(figSize=figSize)
        if refTime == None:
            refTime = self.baseTime + np.min(self.relative_seconds)
        if secWidth <=0:
            raise Exception("Width <= 0")
        # plot all events
        if addSection==True:
            alon = alonlat[0]; alat = alonlat[1]
            blon = blonlat[0]; blat = blonlat[1]
            results = in_rectangle(self.locs,alon,alat,blon,blat,secWidth/2)
            jj = np.where(results[:,0]==1)
            if crop == True:
                self.update_keys(jj)
                self.update_relative_seconds(jj)
                self.update_locs(jj)
                self.update_dict()
                
        if cmap == None:
            plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    (self.locs[:,3]+2)*sizeRatio,
                    edgecolors = edgeColor,
                    facecolors='none',
                    marker='o',
                    alpha=1)
        else:
            shift_seconds = refTime - self.baseTime
            times_plot = self.relative_seconds-shift_seconds
            if unit=="day":
                times_plot = times_plot/(24*60*60)
            elif unit=="hour":
                times_plot = times_plot/(60*60)
            elif unit=="minute":
                times_plot = times_plot/60
            im = plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    c=times_plot,
                    edgecolors=edgeColor,
                    linewidths=edgeWidth,
                    s=(self.locs[:,3]+2)*sizeRatio,
                    cmap = cmap,
                    vmin = vmin,
                    vmax = vmax,
                    marker='o',
                    alpha=1)
            cb = plt.colorbar(im)
            cb.set_label(unit)

        # plot large events
        if impMag != None:
            kk = np.where(self.locs[:,3]>=impMag)
            if len(kk)>0:                 
                imp = plt.scatter(self.locs[kk,0],
                        self.locs[kk,1],
                        (self.locs[kk,3]+2)*sizeRatio*20,
                        edgecolors ='black',
                        facecolors='red',
                        marker='*',
                        alpha=1)
                plt.legend([imp],[f"M$\geq${format(impMag,'4.1f')}"])
        
        if addSection == True: # draw cross-section plot
            a1lon,a1lat,b1lon,b1lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=secWidth/2,
                                                   direction="right")
            a2lon,a2lat,b2lon,b2lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=secWidth/2,
                                                   direction="left")
            plt.plot([a1lon,b1lon,b2lon,a2lon,a1lon],
                     [a1lat,b1lat,b2lat,a2lat,a1lat],
                     linestyle='--',
                     c='darkred')
            plt.plot([alonlat[0],blonlat[0]],[alonlat[1],blonlat[1]],c='darkred')
        # adjust plot parameters
        if len(xlim) != 0:
            plt.xlim(xlim)
        if len(ylim) != 0: 
            plt.ylim(ylim)
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        plt.gca().set_aspect("equal")
        
    def vplot(self,
              alonlat,
              blonlat,
              width=0.1,
              edgecolor='grey',
              edgewidth=0.1,
              depmin=0,
              depmax=10,
              size_ratio=1,
              impMag=None,
              reftime = UTCDateTime(2019,3,1),
              cmap=None,
              vmin=0,
              vmax=1,
              unit="day",
              aspect="auto",
              opposite=1,
              drawFig=True,
              drawCb=False,
              drawImpLg=True,
              ):
        """
        Description

        Parameters
        |   alonlat: the [lon,lat] of the section start point 'a'
        |   blonlat: the [lon,lat] of the section end point 'b'
        |     width: width of section in degree
        |    depmin: minimum depth in km, e.g. 0  km
        |    depmax: maximum depth in km, e.g. 10 km
        |   figsize: e.g. (5,5). Default None means auto set by matplotlib
        | edgecolor: earthquake marker(circle) edgecolor
        |   impMag: important magnitude. Magnitude larger than this level will be 
        |            highlighted
        |  reftime: reference time in UTCDateTime used to constrain colormap, if
        |            no colormap provided, seismicity will plotted by default
        |      cmap: colormap, check 'matplotlib' for more detail.
        | vmin,vmax: the minimum and maximum value for colormap
        |      unit: "day","hour", or "second" for vmin and vmax
        |    aspect: aspect ratio setting. Check for plt.gca().set_aspect for detail
        """
        alon = alonlat[0]; alat = alonlat[1]
        blon = blonlat[0]; blat = blonlat[1]
        length_m,_,_ = gps2dist_azimuth(alat,alon,blat,blon)
        length_km = length_m/1000
        results = in_rectangle(self.locs,alon,alat,blon,blat,width/2)
        jj = np.where(results[:,0]>0)
        self.vplotjj = jj
        self.vxy=np.zeros((jj[0].shape[-1],2))
        self.vkeys=np.zeros((jj[0].shape[-1],))
        self.vkeys = self.keys[jj].ravel()
        self.vxy[:,0] = results[jj,1]
        self.vxy[:,1] = self.locs[jj,2]
        if drawFig:
            if cmap==None:
                plt.scatter(results[jj,1]*opposite,
                        self.locs[jj,2],
                        marker='o',
                        edgecolors = edgecolor,
                        linewidths = edgewidth,
                        facecolors='none',
                        s=(self.locs[jj,3]+2)*size_ratio*5)
            else:
                shift_seconds = reftime - self.baseTime
                times_plot = self.relative_seconds[jj]-shift_seconds
                if unit=="day":
                    times_plot = times_plot/(24*60*60)
                elif unit=="hour":
                    times_plot = times_plot/(60*60)
                elif unit=="minute":
                    times_plot = times_plot/60
                im = plt.scatter(results[jj,1]*opposite,
                        self.locs[jj,2],
                        c=times_plot,
                        s=(self.locs[jj,3]+2)*size_ratio*5,
                        edgecolors=edgecolor,
                        linewidths = edgewidth,
                        cmap = cmap,
                        vmin = vmin,
                        vmax = vmax,
                        marker='o',
                        alpha=1)
                if drawCb:
                    cb = plt.colorbar(im)
                    cb.set_label(unit)

            tmplocs = self.locs[jj]
            tmpresults = results[jj]
            if impMag != None:
                kk = np.where(tmplocs[:,3]>=impMag)
                if len(kk)>0:
                    self.impScatter = plt.scatter(tmpresults[kk,1]*opposite,
                            tmplocs[kk,2],
                            (tmplocs[kk,3]+2)*size_ratio*50,
                            edgecolors =edgecolor,
                            facecolors='red',
                            marker='*',
                            alpha=1)
                    if drawImpLg:
                        plt.legend([self.impScatter],[f"M$\geq${format(impMag,'4.1f')}"])
        
            plt.ylim([depmax,depmin])
            plt.xlim([0,length_km])
            plt.xlabel("distance (km)")
            plt.ylabel("depth (km)")
            plt.gca().set_aspect(aspect)
    
    def MT_plot(self,
                xlim=[],
                ylim=[0,5],
                unit="day",
                refTime=None,
                cmap=None,
                vmin=0,
                vmax=1,
                figSize=(10,5),
                pltShow=True):
        """
        unit: 'day','hour' or 'second'
        """
        # ----------------- initiate -----------------------------------
        assert unit in ['day','hour','second']
        fig,ax = plt.subplots(1,figsize=figSize)
        if unit == "day":
            denominator = (24*60*60)
            plt.xlabel("Time (day)")
        elif unit == "hour":
            denominator = (60*60)
            plt.xlabel("Time (hour)")
        elif unit == "second":
            denominator = 1
            plt.xlabel("Time (second)")
        if refTime == None:
            refTime = self.baseTime + np.min(self.relative_seconds)
        print("[Class Catalog] Reference time is: ", refTime)
        if len(xlim) == 0:
            tbtime = self.baseTime + np.min(self.relative_seconds)
            tetime = self.baseTime + np.max(self.relative_seconds)
        if len(xlim) == 1:
            tbtime = refTime+xlim[0]*denominator
            tetime = self.baseTime + np.max(self.relative_seconds)
        if len(xlim) == 2:
            tbtime = refTime+xlim[0]*denominator
            tetime = refTime+xlim[1]*denominator

        #----------------- processing ---------------------------------
        for key in self.keys:
            etime = self.dict[key][4]
            if etime< tbtime or etime>tetime:
                continue
            emag = self.dict[key][3]
            diff_seconds = etime - refTime
            diff_x = diff_seconds/denominator
            if cmap == None:
                plt.plot([diff_x,diff_x],[ylim[0],emag],c='grey')
            else:
                plt.plot([diff_x,diff_x],[ylim[0],emag],color=cmap((diff_x-vmin)/(vmax-vmin)))
            plt.plot([diff_x],emag,'x',c='k')
        plt.ylim(ylim)
        if len(xlim)>0:
            plt.xlim(xlim)
        else:
            plt.xlim(left = (tbtime-refTime)/denominator-0.1,right = (tetime-refTime)/denominator+0.1)
        plt.ylabel("Magnitude")

        if pltShow == True:
            plt.show()

    def dep_dist_plot(self,refid=None,
                  refloc = [],
                  reftime=None,
                  xlim=[],
                  deplim=[100,-4],
                  distlim=[],
                  unit="day",
                  cmap=None,
                  vmin=0,
                  vmax=1,
                  figsize=(8,6),
                  plt_show=True):
        #------------- initiation --------------------
        fig,axs = plt.subplots(2,1,figsize=figsize)
        if unit == "day":
            denominator = (24*60*60)
            plt.xlabel("Time (day)")
        elif unit == "hour":
            denominator = (60*60)
            plt.xlabel("Time (hour)")
        elif unit == "second":
            denominator = 1
            plt.xlabel("Time (second)")
        if len(xlim)>0:
            axs[0].set_xlim(xlim)
            axs[1].set_xlim(xlim)
        axs[0].set_ylabel("Depth (km)")
        axs[1].set_ylabel("3D-dist (km)")
        axs[0].set_ylim(deplim)
        if len(distlim)>0:
            axs[1].set_ylim(distlim)
        axs[0].grid(axis="y")
        axs[1].grid(axis="y")
        if refid != None:
            reflon = self.dict[refid][0]
            reflat = self.dict[refid][1]
            refdep = self.dict[refid][2]
        if len(refloc)>0:
            reflon,reflat,refdep = refloc
        if reftime == None:
            reftime = self.baseTime + np.min(self.relative_seconds)
        '''
        for evid in self.keys:
            etime =self.dict[evid][4]
            elon = self.dict[evid][0]
            elat = self.dict[evid][1]
            edep = self.dict[evid][2]
            emag = self.dict[evid][3]
            diff_x = (etime-reftime)/denominator
            dist,_,_ = gps2dist_azimuth(elat,elon,reflat,reflon)
            d3dist = np.sqrt((dist/1000)**2+(edep-refdep)**2)
            if cmap==None:
                axs[0].scatter(diff_x,edep,s=(emag+2)*5,marker='o',c='k')
                axs[1].scatter(diff_x,d3dist,s=(emag+2)*5,marker='o',c='k')
            else:
                axs[0].scatter(diff_x,edep,s=(emag+2)*5,marker='o',color=cmap((diff_x-vmin)/(vmax-vmin)))
                axs[1].scatter(diff_x,d3dist,s=(emag+2)*5,marker='o',color=cmap((diff_x-vmin)/(vmax-vmin)))
        '''
        diff_xs = []
        edeps = []
        emags = []
        d3dists = []
        for evid in self.keys:                                                                                                                                                                                     
            etime =self.dict[evid][4]
            elon = self.dict[evid][0]
            elat = self.dict[evid][1]
            edep = self.dict[evid][2]
            emag = self.dict[evid][3]
            diff_x = (etime-reftime)/denominator
            dist,_,_ = gps2dist_azimuth(elat,elon,reflat,reflon)
            d3dist = np.sqrt((dist/1000)**2+(edep-refdep)**2)

            diff_xs.append(diff_x)
            edeps.append(edep)
            emags.append(emag)
            d3dists.append(d3dist)
        emags = np.array(emags)
        if cmap==None:
            axs[0].scatter(diff_xs,edeps,s=(emags+2)*5,marker='o',c='k')
            axs[1].scatter(diff_xs,d3dists,s=(emags+2)*5,marker='o',c='k')
        else:
            axs[0].scatter(diff_xs,edeps,s=(emags+2)*5,marker='o',color=cmap((diff_xs-vmin)/(vmax-vmin)))
            axs[1].scatter(diff_xs,d3dists,s=(emags+2)*5,marker='o',color=cmap((diff_xs-vmin)/(vmax-vmin)))

        plt.tight_layout()
        if plt_show == True:
            plt.show()
        
    def depth_hist(self,mag_threshold=-9,depthmin=0,depthmax=10,gap=0.5,ax=None,pltShow=True,color='grey',edgecolor='k'):
        bins=np.arange(depthmin,depthmax,gap)
        if ax==None:
            fig,ax = plt.subplots(1,1,figsize=(6,8))
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Depth (km)",fontsize=14)
        ax.set_xlabel("Event Qty",fontsize=14)
        kk = np.where(self.locs[:,3]>=mag_threshold)
        hist,bins = np.histogram(self.locs[:,2],bins=bins)
        ax.barh(bins[:-1]+gap/2,hist,height=gap,color=color,edgecolor=edgecolor)
        ax.set_ylim([depthmax,depthmin])
        if pltShow:
            plt.show()
        else:
            return ax
        
    def day_hist(self,
                 refTime=UTCDateTime(2019,1,1,0,0,0),
                 xlim=[],
                 ylim=[],
                 color='b',
                 edgeColor='k',
                 plotMonths=True,
                 figSize=None,
                 pltShow = True):
        """
        Plot events by day-quantity in a histogram plot.
        Parameters:
            -refTime: Reference time for plot
        """
        ref_list = []
        time_list = []
        for key in self.dict.keys():
            _,_,_,_,etime = self.dict[key]
            ref_list.append((etime-refTime)/(24*60*60))

        min_day=floor(min(ref_list))
        max_day=ceil(max(ref_list))
        bins = np.linspace(min_day,max_day,max_day-min_day+1)
        if figsize==None:
            figsize=(8,4)
        fig1 = plt.figure(1,figsize=figsize)
        ax1 = plt.subplot(1,1,1)
        ax1.hist(ref_list,bins,color=color,edgecolor=edgecolor)
        # The bottom x-axis is in days
        ax1.set_xlim([0,max_day])
        # The top x-axis marks year and month in YYYYMM
        tick_list_1 = [] # Store the position number
        tick_list_2 = [] # Store the tick text
        ref_year = refTime.year
        ref_month = refTime.month
        ref_day = refTime.day
        if ref_day == 1:
            tick_list_1.append(0)
            tick_list_2.append(str(ref_year)+str(ref_month).zfill(2))
        status = True # Start to loop month by month
        loop_time = UTCDateTime(ref_year,ref_month,1) # Initiate loop time
        step = 32 #32 > 31. Make sure each step pass to next month
        while status==True:
            loop_time = loop_time + step*24*60*60
            tmp_year = loop_time.year
            tmp_month = loop_time.month
            loop_time = UTCDateTime(tmp_year,tmp_month,1)
            diff_days = (loop_time - refTime)/(24*60*60)
            if diff_days > (max_day):
                status=False
            else:
                tick_list_1.append(diff_days)
                tick_list_2.append((str(tmp_month).zfill(2)))
        if plotMonths:
            ax2 = ax1.twiny()
            ax2.set_xlim([0,max_day])
            ax2.plot(0,0,'k.')
            plt.xticks(tick_list_1,tick_list_2)
            ax2.set_xlabel("date")
        if xlim!=[]:
            ax1.set_xlim(xlim)
            if plotMonths:
                ax2.set_xlim(xlim)
        if ylim!=[]:
            plt.ylim(ylim)
        ax1.set_xlabel("Time, days")
        ax1.set_ylabel("event quantity")

        if pltShow == True:
            plt.show()

    def diffusion_plot(self,refid=None,refloc=[],diff_cfs=[],unit="day",xlim=[],ylim=[]):
        '''
        Parameters:
        refid: reference event id
        refloc: [lon,lat], reference site longitude and latitude, if not provided, use refid
        diff_cfs: diffusion coefficient list, this will draw corresponding lines on the map
        '''
        #from seisloc.plot import diffusion_plot
        
        if refid==None and refloc==[]:
            raise Exception("refid or refloc should be proivded")
        if refloc==[]:
            refloc=[self.dict[refid][0],self.dict[refid][1]]
        dist_list = np.zeros((len(self.keys),1))
        day_list = np.zeros((len(self.keys),1))
        mag_list = np.zeros((len(self.keys),1))
        for i in range(len(self.keys)):
            dist,_,_ = gps2dist_azimuth(self.locs[i,1],\
                                self.locs[i,0],\
                                refloc[1],\
                                refloc[0])
            day_list[i,0] = (self.relative_seconds[i]-np.min(self.relative_seconds))/(24*60*60)
            mag_list[i,0] = self.locs[i,3]
            dist_list[i,0] = dist

        fig1 = plt.figure(1)
        ax1 = plt.subplot(1,1,1)
        if unit=="day":
            x_list = day_list
            plt.xlabel("Time (day)",fontsize=16)
        elif unit == "hour":
            x_list = day_list*24
            plt.xlabel("Time (hour)",fontsize=16)
        else:
            raise Exception("Unit error: 'day' or 'hour'")
        ax1.set_ylabel("Distance (m)",fontsize=16)
        ax1.scatter(x_list,dist_list,(mag_list+2)*3,c='k')
        ax1.set_xlim([0,np.max(x_list)])

        diff_lines = []
        if isinstance(diff_cfs,int) or isinstance(diff_cfs,float):
            diff_cfs=[diff_cfs]
        for diff_cf in diff_cfs:
            if unit=="day":
                x = np.linspace(0,np.max(x_list),int(np.max(x_list)*20)+1)
                y = np.sqrt(4*np.pi*diff_cf*x*24*60*60)
            elif unit=="hour":
                x = np.linspace(0,np.max(x_list),int(np.max(x_list)*20)+1)
                y = np.sqrt(4*np.pi*diff_cf*x*60*60)
            diff_line, = plt.plot(x,y)
            diff_lines.append(diff_line)

        plt.legend(diff_lines,diff_cfs,title="Diffusion Coefficient $m^2/s$")
        if len(xlim)>0:
            plt.xlim(xlim)
        if len(ylim)>0:
            plt.ylim(ylim)
        else:
            plt.ylim(bottom=0)

    def animation(self,
                  incre_hour=2,
                  mb_time=None,
                  me_time=None,
                  xlim=[],
                  ylim=[],
                  geopara=None,
                  cmap=None,
                  vmin=None,
                  vmax=None
                  ):
        """
        Generate gif animation file
        increment: Time increased for each plot. Unit: hour
        """
        # Remove previous results
        try:
            shutil.rmtree("dd_animation")
        except:
            pass
        os.makedirs("dd_animation")
        if xlim == []:
            xlim = [np.min(self.locs[:,0]),np.max(self.locs[:,0])]
        if ylim == []:
            ylim = [np.min(self.locs[:,1]),np.max(self.locs[:,1])]    
        min_time = self.baseTime+np.min(self.relative_seconds)
        max_time = self.baseTime+np.max(self.relative_seconds)
        
        if mb_time == None:
            mb_time = min_time
        if me_time == None:
            me_time = max_time
        print("[Class Catalog] Movie start time is: ",mb_time)
        print("[Class Catalog]   Movie end time is: ",me_time)
        #---------- cmap -------------------------
        if vmin == None:
            vmin = 0
        if vmax == None:
            vmax = (max_time - min_time)/(24*60*60)

        inc_second = incre_hour*60*60 # Time increment
        loop_time = mb_time
        count = 1
        reftime = mb_time
        while loop_time <= me_time:
            fig = plt.figure(1,figsize=(8,8))
            ax1 = fig.add_subplot(111)
            ax1.set_xlim(xlim)
            ax1.set_ylim(ylim)
            ax1.set_xlabel("Lon(degree)",fontsize=18)
            ax1.set_ylabel("Lat(degree)",fontsize=18)
            ax1.set_title(f"{str(loop_time)[:19]}",fontsize=16)
            if geopara != None:
                # --------------- The Z201H02 pad ----------------------------
                padZ201H02 = geopara.wellpad("Z201H02",well_edgecolor="purple")
                padZ201H04 = geopara.wellpad("Z201H04",well_edgecolor="purple")
                ax1.add_collection(padZ201H02)
                ax1.add_collection(padZ201H04)
                # ----------------Molin faults--------------------------------
                ml_fault = np.array(geopara.dict['ml_fault'])
                ML_fault,=ax1.plot(ml_fault[:,0],ml_fault[:,1],'r-')
                # ----------------Zigong faults-------------------------------
                for key in geopara.dict['zg_faults']:
                    array = np.array(geopara.dict['zg_faults'][key])
                    ax1.plot(array[:,0],array[:,1],'k-',label='Faults')
                #----------------------Wells-----------------------------------
                wells = np.array(geopara.dict['wells'])
                well_lons=[]; well_lats=[]
                for well in geopara.dict['wells']:
                    well_lons.append(well[0])
                    well_lats.append(well[1])
                s_well,=ax1.plot(well_lons,well_lats,'s',c='#1f77b4',markerfacecolor='white',mew=2,markersize=12)
                #----------------------Stations--------------------------------
                sta_lons=[]; sta_lats=[]
                for sta in geopara.dict['sta_locs']:
                    if sta[3]=="SC":
                        sta_lons.append(sta[0]); sta_lats.append(sta[1])
                if len(sta_lons)>0:
                    s_sta=ax1.scatter(sta_lons,sta_lats,marker='^',c='cyan',s=120,edgecolor='k',label='Stations')
            #------------- Events -----------------------------------------
            eve_arr = []
            rela_days = []
            for i,second in enumerate(self.relative_seconds):
                e_time = self.baseTime + second
                if e_time<(loop_time+inc_second/2) and e_time>mb_time:
                    eve_arr.append(self.locs[i,:4])
                    rela_days.append((e_time-reftime)/(24*60*60))
            eve_arr = np.array(eve_arr)
            rela_days = np.array(rela_days)

            if len(eve_arr)>0:
                if cmap != None:
                    s_eve=ax1.scatter(eve_arr[:,0],
                                          eve_arr[:,1],
                                          s=(eve_arr[:,3]+2)*5,
                                          c=rela_days,
                                          cmap=cmap,
                                          vmin=vmin,
                                          vmax=vmax,
                                          label="Events")
                else:
                    s_eve=ax1.scatter(eve_arr[:,0],
                                          eve_arr[:,1],
                                          s=(eve_arr[:,3]+2)*5,
                                          c='k',
                                          label="Events")
            if geopara != None:
                plt.legend([s_well,s_sta,s_eve],\
                       ["Platform",'Station',"Seismicity"],\
                       loc='upper right',
                       fontsize=16)  
            ##------------- save results --------------------------------------------
            plt.savefig(f"dd_animation/{str(count).zfill(3)}.png")
            loop_time = loop_time + inc_second
            count+=1
            plt.close()
        #-------------------- gif -----------------------------
        imgs = []
        for i in range(1,count):
            pic_name = f'dd_animation/{str(i).zfill(3)}.png'
            tmp = Image.open(pic_name)
            imgs.append(tmp)
        imgs[0].save("dd_animation.gif",save_all=True,append_images=imgs,duration=10)
    
    def intervals_plot(self,interval=1,method='month',
                       xlim=[],ylim=[],
                       columns=4,subplotsize=(3,3),
                       marker='o',ms=1,
                       wspace=None,hspace=None):

        from seisloc.plot import intervals_plot
        plt.close()
        axs = intervals_plot(xys=self.locs[:,:2],
                rela_secs=self.relative_seconds,
                reftime=self.baseTime,
                interval=interval,method=method,
                xlim=xlim,ylim=ylim,
                columns=columns,subplotsize=subplotsize,
                marker=marker,ms=ms,
                wspace=wspace,hspace=hspace)

        return axs

    def depths_plot(self,
                deplim=[0,10],interval=1,
                xlim=[],ylim=[],
                columns=4,subplotsize=(3,3),
                marker='o',ms=1,color='k',
                zorder=0,
                wspace=None,hspace=None):
        from seisloc.plot import depths_plot
        axs = depths_plot(xyz=self.locs[:,:3],
                          deplim=deplim,interval=interval,
                          xlim=xlim,ylim=ylim,
                          columns=columns,subplotsize=subplotsize,
                          marker=marker,ms=ms,color=color,
                          zorder=zorder,
                          wspace=None,hspace=None)
        return axs
    def sum_count_Mo(self,starttime,endtime,outFile='sum_count_Mo.txt',mode='day'):
        self.dict_count_Mo = sum_count_Mo(self,starttime,endtime)
        write_sum_count_Mo(self.dict_count_Mo,outFile,mode)

    
    

    def write_info(self,fileName=None,refTime=None,disp=False):
        if fileName==None:
            nowTime = UTCDateTime.now()
            fileName = "Catalog_"+nowTime.strftime("%Y%m%d%H%M%S")+".txt"
        if refTime == None:
            refTime = self.baseTime
        print("[Class Catalog] The reference time is: ",refTime)
        f = open(fileName,'w')
        for key in self.keys:
            lon = self.dict[key][0]
            lat = self.dict[key][1]
            dep = self.dict[key][2]
            if dep>=6800:
                print(f"[Warning] Depth of event {key} exceeds 6800, are you sure? ")
            mag = self.dict[key][3]
            etime = self.dict[key][4]
            relative_days = (self.dict[key][4]-refTime)/(24*60*60)
            _key = format(key,'8d')
            _lon = format(lon,'12.6f')
            _lat = format(lat,'11.6f')
            _dep = format(dep,'8.2f')
            _mag = format(mag,'5.1f')
            _relative_days = format(relative_days,'16.8f')
            line = _key+_lon+_lat+_dep+_mag+_relative_days+" "+str(etime)
            f.write(line+"\n")
            if disp == True:
                print("[Class Catalog] "+line)
        print("[Class Catalog] Catalog information write into: ",fileName)
        f.close()

    def cata2fdsn(self,author="Hardy",catalog="SC",
                  cont="SC",contID="01",magtype="ML",
                  magauthor="SC Agency",elocname="SC",out_file='cata.fdsn'):

        cata2fdsn(self,author=author,catalog=catalog,
                  cont=cont,contID=contID,magtype=magtype,
                  magauthor=magauthor,elocname=elocname,out_file=out_file)

    def copy(self):
        return copy.deepcopy(self)
    def merge(self,dd2):
        for key in dd2.keys:
            if key in self.keys:
                raise Exception(f"Key error {key}. Please avoid using the same key value")
            self.dict[key]=dd2.dict[key]
        self.init()

    def __repr__(self):
        _qty = f"HypoDD relocation catalog with {len(self.dict.keys())} events\n"
        _time= f"     Time range is: {self.baseTime+np.min(self.relative_seconds)} to {self.baseTime+np.max(self.relative_seconds)}\n"
        _mag = f" Magnitue range is: {format(np.min(self.locs[:,3]),'4.1f')} to {format(np.max(self.locs[:,3]),'4.1f')}\n"
        _lon = f"Longitude range is: {format(np.min(self.locs[:,0]),'8.3f')} to {format(np.max(self.locs[:,0]),'8.3f')}\n"
        _lat = f" Latitude range is: {format(np.min(self.locs[:,1]),'7.3f')} to {format(np.max(self.locs[:,1]),'7.3f')}\n"
        _dep = f"    Depth range is: {format(np.min(self.locs[:,2]),'4.1f')} to {format(np.max(self.locs[:,2]),'4.1f')}\n"
        return _qty+_time+_mag+_lon+_lat+_dep

    def __getitem__(self,key):
        return self.dict[key]

def hypoinv2Catalog(inv):
    """
    Convert Hypoinv class to Catalog class
    """
    inv_dict={}
    for key in inv.dict_evid.keys():
        inv_dict[key] = inv.dict_evid[key][1:5]
        _time = inv.dict_evid[key][0]
        etime = UTCDateTime.strptime(_time,"%Y%m%d%H%M%S%f")
        inv_dict[key].append(etime)
    inv_cata = Catalog(locFile=None)
    inv_cata.dict = inv_dict
    inv_cata.init()
    return inv_cata

def dtcc_otc(dtcc_old,inv_old_file,inv_new_file):
    """
    Conduct dtcc origin time correction in updated out.sum
    The output file is a new file with suffix .otc after the input file
    """
    #------------ load data ---------------------
    with open(dtcc_old,'r') as f:
        lines = f.readlines()
    inv_old = Hypoinv(inv_old_file)
    inv_new = Hypoinv(inv_new_file)
    
    #------------ processing --------------------
    f = open("dtcc.otc",'w')
    for line in tqdm(lines):
        line = line.rstrip()
        if line[0] == "#":     # event pair line
            status = True
            _,_id1,_id2,_otc = line.split()
            id1 = int(_id1)
            id2 = int(_id2)
            otc = float(_otc)
            # ------------- read old event time ---------------
            _et1_old = inv_old[id1][0]
            et1_old = UTCDateTime.strptime(_et1_old,'%Y%m%d%H%M%S%f')
            _et2_old = inv_old[id2][0]
            et2_old = UTCDateTime.strptime(_et2_old,'%Y%m%d%H%M%S%f')
            # ------------- read new event time ---------------
            try:
                _et1_new = inv_new[id1][0]
            except:
                status = False       # events not included in the new set
                continue
            et1_new = UTCDateTime.strptime(_et1_new,'%Y%m%d%H%M%S%f')
            try:
                _et2_new = inv_new[id2][0]
            except:
                status = False       # events not included in the new set
                continue
            et2_new = UTCDateTime.strptime(_et2_new,'%Y%m%d%H%M%S%f')
            # ------------- calculate otc --------------------
            det1 = et1_new - et1_old
            det2 = et2_new - et2_old
            otc = otc+(det1-det2)
            # ------------- prepare writting -----------------
            line = line[:14]+format(otc,'.2f')
        if status == True:
            f.write(line+"\n")
    f.close()

def read_txt_cata(cataPth,verbose=1):
    """
    This function could be replaced by Catalog(cataPth,format="cata") [recommend]
    """
    with open(cataPth,'r') as f:
        edict={}
        for line in f:
            line = line.strip()
            _evid,_evlo,_evla,_evdp,_emag,_eday,_etime=line.split()
            edict[int(_evid)] = [float(_evlo),float(_evla),float(_evdp),float(_emag),UTCDateTime(_etime)]
    cata = Catalog(locFile=None,verbose=verbose)
    cata.dict = edict
    cata.init()
    return cata



import re
import os
import shutil
import random
from obspy import UTCDateTime
from tqdm import tqdm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import multiprocessing as mp
import time
import subprocess
from seisloc.geometry import spherical_rotate
from seisloc.utils import readfile,writefile
import pandas as pd

def load_PC(catalog="/home/zijinping/Desktop/projects/wy_eq/2018_2019_PC/2018_2019_hypoDD.reloc",
            start_evid=300000):
    """
    Read in Pengcheng's earthquake catalog
    """
    evid = start_evid+1
    catalog_dict = {}
    cont = []
    with open(catalog,'r') as f:
        cont = f.readlines()
    for line in cont:
        _time,_,_lat,_lon,_dep,_mag, = re.split(" +",line.rstrip())[:6]
        etime = UTCDateTime.strptime(_time,'%Y%m%d%H%M%S')
        catalog_dict[evid] = [float(_lon),float(_lat),float(_dep),float(_mag),etime]
        evid += 1
    return catalog_dict

def load_CEDC(catalog="/home/zijinping/Dropbox/resources/catalog/CEDC/20090101_20201231.txt",start_evid=100000):
    """
    Read in China Earthquake Data Center catalog
    """
    evid = start_evid+1
    catalog_dict = {}
    cont = []
    with open(catalog,'r') as f:
        cont = f.readlines()
    for line in cont:
        _time,_lon,_lat,_dep,M,_mag,_,_ = re.split(",",line.rstrip())
        #print(_time,_lon,_lat,_dep,_mag)
        _date,_hr_min = re.split(" ",_time)
        _yr,_mo,_dy = re.split("\/",_date)
        _hr,_min = re.split(":",_hr_min)
        yr = int(_yr); mo = int(_mo); dy = int(_dy)
        hr = int(_hr); minute = int(_min)
        etime = UTCDateTime(yr,mo,dy,hr,minute,0)
        catalog_dict[evid] = [float(_lon),float(_lat),float(_dep),float(_mag),etime]
        evid += 1
    return catalog_dict


def dd_event_sel(evid_list=[],event_dat="event.dat",event_sel="event.sel"):
    '''
    select events in the "event.dat" file and output them into
    the "event.sel" file by the event ID list provided
    '''
    content = []
    # Read in data
    with open(event_dat,'r') as f:
        for line in f:
            line = line.rstrip()
            evid = int(line[-8:])
            if evid in evid_list:
                content.append(line)
    f.close()

    # Output into target file
    with open(event_sel,'w') as f:
        for line in content:
            f.write(line+"\n")
    f.close()

def dtct_sel(evid_list,input_file):
    """
    Output clean dtct file with event id list provided
    """
    out_file = input_file+".sel"
    out_cont = []
    record_status = False # Set initiate value
    with open(input_file,'r') as f:
        for line in f:
            if line[0]=="#": # pair line
                print(re.split(" +",line.rstrip()))
                _,ID1,ID2 = re.split(" +",line.rstrip())
                ID1 = int(ID1)
                ID2 = int(ID2)
                if (ID1 in evid_list) and (ID2 in evid_list):
                    record_status = True
                    out_cont.append(line.rstrip())
                else:
                    record_status = False
            elif record_status==True:
                out_cont.append(line.rstrip())
    f.close()

    with open(out_file,'w') as f:
        for line in out_cont:
            f.write(line+"\n")


def hypoDD_rmdup(in_file="total_hypoDD.reloc"):
    """
    remove duplicated events and take mean values
    """
    count=0
    evid_list = []
    evid_mapper = {}
    with open(in_file,'r') as f:
        for line in f:
            evid = int(line[0:11])
            try:
                evid_mapper[evid].append(line)
                count += 1
            except:
                evid_mapper[evid]=[line]
    f.close()

    evid_list = list(evid_mapper)
    evid_list.sort()

    log_record = []
    f=open(in_file+".rm",'w')
    f.close()
    for evid in evid_list:
        if len(evid_mapper[evid])>1:
            lon_list = []
            lat_list = []
            dep_list = []
            for i in range(len(evid_mapper[evid])):
                lon = float(evid_mapper[evid][i][22:32])
                lon_list.append(lon)
                lat = float(evid_mapper[evid][i][11:20])
                lat_list.append(lat)
                dep = float(evid_mapper[evid][i][36:42])
                dep_list.append(dep)
            lon_mean = np.mean(lon_list)
            lat_mean = np.mean(lat_list)
            dep_mean = np.mean(dep_list)
            lon_str = format(lon_mean,'10.6f')
            lat_str = format(lat_mean,'9.6f')
            dep_str = format(dep_mean,'6.3f')
            log_record.append([evid,lon_str,lat_str,dep_str])
            firstr = evid_mapper[evid][0]  #Use the first record as template
            outstr = firstr.replace(line[22:32],lon_str,1) #Replace lon
            outstr = outstr.replace(line[11:20],lat_str,1) #Replace lat
            outstr = outstr.replace(line[22:32],dep_str,1) #Replace dep
            with open("total_hypoDD.reloc.rm",'a') as f:
                f.write(outstr)
            f.close()
        else:
            with open(in_file+".rm",'a') as f:
                f.write(evid_mapper[evid][0])
            f.close()
    print(log_record)

def hypoDD_ref_days(reloc_file,ref_time,shift_hours=0):
    """
    Add one column to the last of hypoDD files, calculate the length of time 
    between the referece time and the event time in days.
    The output is a file with the same title with reloc_file and add ".add" as
    suffix.

    Parameters
    ----------
     reloc_file: The hypoDD relocation file.
       ref_time: Reference time in UTCDateTime format
    shift_hours: Used when event time is not in UTC time zone
    """

    new_add=[]
    with open(reloc_file,"r") as f:
        for line in f:
            year = int(re.split(" +",line)[11])
            month = int(re.split(" +",line)[12])
            day = int(re.split(" +",line)[13])
            hour = int(re.split(" +",line)[14])
            minute = int(re.split(" +",line)[15])
            seconds = float(re.split(" +",line)[16])
            eve_time = UTCDateTime(year,month,day,hour,minute)+seconds
            days = (eve_time - ref_time)*1.0/(24*60*60)
            new_line=line[:-1]+" "+format(days,'4.2f')
            new_add.append(new_line)
    f.close()
    with open(reloc_file+".add","w") as f:
        for line in new_add:
            f.write(line+"\n")
    f.close()

def compare_DD(dd1_path,dd2_path):
    dd1,_ = load_DD(dd1_path)
    dd2,_ = load_DD(dd2_path)
    f = open("dd_diff.dat",'w')
    for key in dd1:
        try: 
            lon1 = dd1[key][0] 
            lon2 = dd2[key][0] 
            lat1 = dd1[key][1] 
            lat2 = dd2[key][1] 
            dep1 = dd1[key][2] 
            dep2 = dd2[key][2]
            print("dx,dy,dz:",abs(lon1-lon2)*111*1000,abs(lat1-lat2)*111*1000,abs(dep1-dep2)*1000)
            f.write(f"{abs(lon1-lon2)*111*1000} {abs(lat1-lat2)*111*1000} {abs(dep1-dep2)*1000}\n")
        except: 
            pass 
    f.close()

def hypoDD_mag_mapper(reloc_file,out_sum,magcolumn_index=128):
    """    
    The output of hypoDD doesn't contain magnitude information.
    This function reads magnitude information from *.sum file, which is the
    output of hyperinverse and provide to hypoDD file.
           
    The results will cover the input reloc_fiie
    """    
           
    #get the magnitude dictionary
    event_mag_list = {}
    with open(out_sum,"r") as f_obj:
        for line in f_obj:
            event_id = int(line[136:146])
            event_mag = int(line[123:126])*0.01
            event_mag_list[event_id]=event_mag
    f_obj.close()
    print(len(event_mag_list.keys()))
    #add in the magnitude

    new_dd = []
    with open(reloc_file,"r") as f_obj:
        for line in f_obj:
            dd_event_id = int(line[0:9])
            dd_event_mag = event_mag_list[dd_event_id]
            new_line=line[:magcolumn_index]+format(dd_event_mag,'5.2f')+line[magcolumn_index+4:]
            new_dd.append(new_line)
    f_obj.close()
    with open(reloc_file+".mag","w") as f_obj:
        for line in new_dd:
            f_obj.write(line)
    f_obj.close()

def load_DD(reloc_file="hypoDD.reloc",shift_hour=0):
    """
    load results of hypoDD
    return eve_dict, df

    Parameters
    ----------
    If the time of results is not in UTC time zone, a time shift might needed.
    For example, Beijing time zone is 8 hours early than UTC time, 8 hours 
    should be deducted so as to be consistent with UTC time.
    """

    eve_dict={}
    columns = ["ID","LAT","LON","DEPTH","X","Y","Z","EX","EY","EZ",\
           "YR","MO","DY","HR","MI","SC","MAG",\
           "NCCP","NCCS","NCTP","NCTS","RCC","RCT","CID"]
    number = 0
    dataset = []
    with open(reloc_file,'r') as f:
        for line in f:
            line = line.rstrip()
            if "*" not in line:
                data = list(map(float,line.split()))
                dataset.append(data)
    dataset = np.array(dataset)
    
    if dataset.shape[1] == 24: # format of hypoDD.reloc file
        columns = ["ID","LAT","LON","DEPTH","X","Y","Z","EX","EY","EZ",\
                   "YR","MO","DY","HR","MI","SC","MAG",\
                   "NCCP","NCCS","NCTP","NCTS","RCC","RCT","CID"]
    if dataset.shape[1] == 25: # format of modified hypoDD.reloc, note the last column
        columns = ["ID","LAT","LON","DEPTH","X","Y","Z","EX","EY","EZ",\
                   "YR","MO","DY","HR","MI","SC","MAG",\
                   "NCCP","NCCS","NCTP","NCTS","RCC","RCT","CID","DAY"]
    if dataset.shape[1] == 18: # format of hypoDD.loc file
        columns = ["ID","LAT","LON","DEPTH","X","Y","Z","EX","EY","EZ",\
                   "YR","MO","DY","HR","MI","SC","MAG","CID"]

    
    for i,data in enumerate(dataset):
        eve_id = data[0]
        eve_lat = data[1]
        eve_lon = data[2]
        eve_dep = data[3]
        eve_mag = data[16]
        eve_time = UTCDateTime(int(data[10]),int(data[11]),int(data[12]),int(data[13]),int(data[14]),0)+data[15] - shift_hour*50*60
        eve_dict[int(eve_id)]=[float(eve_lon),float(eve_lat),float(eve_dep),float(eve_mag),eve_time]

    df = pd.DataFrame(data=dataset,columns=columns)
    return eve_dict,df


def run_dd(base_dir="./",work_dir='hypoDD',inp_file="hypoDD.inp"):
    os.chdir(base_dir)
    os.chdir(work_dir)
    subprocess.run(["hypoDD",inp_file])
    os.chdir(base_dir)

def dd_bootstrap(base_folder="hypoDD",times=10,method="event",samp_ratio=0.75,cores=2):
    """
    Randomly run hypoDD with randomly selected events to show the results variation
    Parameters:
        base_folder: the basement folder which should include the material for hypoDD, 
            including dt.ct, dt.cc,hypoDD.inp, event.dat, station.dd
        times: number of hypoDD runs
        method: "event" means sample events; "phase" means sample phases
        samp_ratio: the ratio of events to be relocated in the run
    """
    # Load in event.dat file
    base_dir = os.getcwd()
    e_dat = []
    if method == "event":
        with open(os.path.join(base_folder,"event.dat"),'r') as f:
            for line in f:
                evid = int(line[84:91])
                e_dat.append(line)
        e_qty = len(e_dat)
        s_qty = int(e_qty*samp_ratio) # sample qty

        tar_folders = []
        tasks = []
        # Prepare the subroutine files
        for i in range(1,times+1):
            tar_folder = base_folder + str(i).zfill(3)
            shutil.copytree(base_folder,tar_folder)
            sel_idxs = random.sample(range(e_qty),s_qty)
            with open(os.path.join(tar_folder,'event.sel'),'w') as f:
                for idx in sel_idxs:
                    f.write(e_dat[idx])
            f.close()
            tar_folders.append(tar_folder)
            tasks.append([base_dir,tar_folder])
            
    if method == "phase":
        dtct = []
        out_dtct = []
        with open(os.path.join(base_folder,"dt.ct"),'r') as f:
            for line in f:
                dtct.append(line)        
        len_dtct = len(dtct)
        
        tar_folders = []
        tasks = []
        # Prepare the subroutine files
        for i in range(1,times+1):
            tar_folder = base_folder + str(i).zfill(3)
            shutil.copytree(base_folder,tar_folder)
            with open(os.path.join(tar_folder,'dt.ct'),'w') as f:
                
                for i,line in enumerate(dtct):
                    if line[0] == "#":    # event line
                        f.write(dtct[i])
                        tmp = []
                        j = i+1
                        while j<len_dtct and dtct[j][0]!="#":
                            tmp.append(dtct[j])
                            j=j+1
                        pha_qty = len(tmp)
                        sample_qty = int(pha_qty*samp_ratio+0.5)
                        sel_idxs = random.sample(range(pha_qty),sample_qty)
                        for idx in sel_idxs:
                            f.write(tmp[idx])


            tar_folders.append(tar_folder)
            tasks.append([base_dir,tar_folder])
            
    pool = mp.Pool(processes=cores)
    rs = pool.starmap_async(run_dd,tasks,chunksize=1)
    while True:
        remaining = rs._number_left
        print(f"Finished {len(tasks)-remaining}/{len(tasks)}",end='\r')
        if(rs.ready()):
            break
        time.sleep(0.5)
    print("\nDone!!!")
    
def bootstrap_summary(times,base_folder="hypoDD"):
    """
    """
    rand_dict = {}
    
    with open(os.path.join(base_folder,"event.dat"),'r') as f:
        for line in f:
            evid = int(line[84:91])
            rand_dict[evid] = []

    
    tar_folders = []
    for i in range(1,times+1):
        tar_folder = base_folder + str(i).zfill(3)
        tar_folders.append(tar_folder)

    # Load hypoDD results
    print("Loading results ...")
    for tar_folder in tar_folders:
        reloc_file = os.path.join(tar_folder,'hypoDD.reloc')
        dd_dict,_ = load_DD(reloc_file)
        for key in dd_dict.keys():
            rand_dict[key].append(dd_dict[key][:4])
    
    print("Write out results ... ")
    f = open("hypoDD.rand",'w')
    for key in rand_dict.keys():
        cont = rand_dict[key]
        if len(cont) == 0:
            continue
        else:
            lon_list = []
            lat_list = []
            dep_list = []
            for tmp in cont:
                lon_list.append(tmp[0])
                lat_list.append(tmp[1])
                dep_list.append(tmp[2])
            record_qty = len(cont)
            mean_lon = np.mean(lon_list)
            mean_lat = np.mean(lat_list)
            mean_dep = np.mean(dep_list)
            std_lon = np.std(lon_list)
            std_lat = np.std(lat_list)
            std_herr = np.sqrt(std_lon**2+std_lat**2)
            std_dep = np.std(dep_list)
            f.write(format(key,'7d')+" "+
                format(mean_lon,'8.4f')+" "+
                format(mean_lat,'7.4f')+" "+
                format(mean_dep*1000,'9.3f')+" "+
                format(std_herr*111.1*1000*2,"9.3f")+" "+
                format(std_dep*1000*2,"8.3f")+" "+
                format(record_qty,'3d')+"\n")
    f.close()

def pha_subset(pha_file,loc_filter,obs_filter=8,out_file=None):
    """
    *.pha file is the input file for hypoDD ph2dt, this function subset the
    pha file by the boundary condition and the minimum observation condition.
    The output file is a file with ".subset" suffix

    Parameters
    ----------
    pha_file: Str. The input file.
    loc_filter: array in format [lon_min, lon_max, lat_min, lat_max]
    obs_filter: The minimum observation
    out_path: file path for the target file
    """

    lon_min, lon_max, lat_min, lat_max = loc_filter
    if out_file == None:
        out_file = pha_file+".subset"
    f = open(out_file,"w")
    f.close()
    pha_content = []
    with open(pha_file,"r") as f:
        for line in f:
            pha_content.append(line.rstrip())
    f.close()
    i = 0
    j = 0
    record_list=[]
    for line in pha_content:
        if line[0]=="#":
            if i>0 and len(record_list) > (obs_filter+1):
                j=j+1
                with open(out_file,"a") as f:
                    for record in record_list:
                        f.write(record+"\n")
                f.close()
                record_list = []
                record_list.append(line)
            else:
                record_list = []
                record_list.append(line)
            i=i+1
            lat = float(re.split(" +",line)[7])
            lon = float(re.split(" +",line)[8])
            if lat>lat_min and lat<lat_max and lon>lon_min and lon<lon_max:
                region_pass = True
            else:
                region_pass = False
        else:
            if region_pass:
                record_list.append(line)
    if i>0 and len(record_list) > (obs_filter+1):
        j=j+1
        with open(out_file,"a") as f:
            for record in record_list:
                f.write(record+"\n")
        f.close()
    print("Event before filtering",i)
    print("Events qty after filtering",j)

def pha_sel(pha_file,e_list=[],remove_net=False):
    '''
    Select phases of events in e_list
    if need to remove net name.
    '''
    out = []   # output
    with open(pha_file,'r') as f:
        for line in f:
            line = line.rstrip()
            if line[0]=="#":
                _evid = re.split(" +",line)[-1]
                evid = int(_evid)
                if evid in e_list or e_list==[]:
                    status = True
                    out.append(line)
                else:
                    status = False
            else:
                if status==True:
                    if remove_net:
                        out.append(line[2:])
                    else:
                        out.append(line)
    f.close()

    with open(pha_file+".sel",'w') as f:
        for line in out:
            f.write(line+'\n')
    f.close()

def inv_dd_compare(inv,dd,keys=[],xlim=[],ylim=[],aspect='auto'):
    inv_locs = []
    dd_locs = []
    for key in keys:
        key = int(key)
        inv_lon = inv[key][1]
        inv_lat = inv[key][2]
        dd_lon = dd[key][0]
        dd_lat = dd[key][1]
        inv_locs.append([inv_lon,inv_lat])
        dd_locs.append([dd_lon,dd_lat])
    inv_locs = np.array(inv_locs)
    dd_locs = np.array(dd_locs)
    plt.plot(inv_locs[:,0],inv_locs[:,1],'kx')
    plt.plot(dd_locs[:,0],dd_locs[:,1],'rv')
    if len(xlim)>0:
        plt.xlim(xlim)
    if len(ylim)>0:
        plt.ylim(ylim)
    plt.gca().set_aspect(aspect)

def rotDD(ddfile,center,rotate):
    """
    Parameters:
    |  center: the rotation center [lon,lat], positive for E and N
    |  rotate: rotate angle in degree, postive for anticlockwise
    """
    cont = readfile(ddfile)
    
    new_cont = []
    for line in cont:
        _evid,_lat,_lon = line.split()[:3]
        lat = float(_lat)
        lon = float(_lon)
        rotated = spherical_rotate([lon,lat],center=center,degree=rotate)
        new_lon = rotated[0,0]
        new_lat = rotated[0,1]
        new_line = line[:9] + format(new_lat,'>11.6f')+format(new_lon,'>12.6f')+line[32:]
        new_cont.append(new_line)
    
    writefile(new_cont,ddfile+".rot")

def subset_dtcc(evids,ccFilePth,saveFilePth="",useStas=[]):
    linesUse = []      
    ccFileName = os.path.basename(ccFilePth)
    with open(ccFilePth,'r') as f:
        for line in tqdm(f):  
            if line[0] == "#":
                record=False
                _,_evid1,_evid2,_ = line.split()
                if int(_evid1) in evids or int(_evid2) in evids:
                    linesUse.append(line)
                    record = True
            else:          
                if record==True:
                    if useStas==[]:
                        linesUse.append(line)
                    else:
                        sta = re.split(" +",line)[0]
                        if sta in useStas:
                            linesUse.append(line)
    if saveFilePth=="":
        saveFilePth=ccFileName+'.subset'
    with open(saveFilePth,'w') as f:
        for line in linesUse:
            f.write(line)

def subset_dd_file(evids,ddFilePth,saveFilePth=""):
    linesUse = []      
    ddFileName = os.path.basename(ddFilePth)
    with open(ddFilePth,'r') as f:
        for line in tqdm(f):  
                _evid = line[:9]
                if int(_evid) in evids:
                    linesUse.append(line)
    if saveFilePth=="":
        saveFilePth=ddFileName+'.subset'
    with open(saveFilePth,'w') as f:
        for line in linesUse:
            f.write(line)
